/***************************************

    This is part of frox: A simple transparent FTP proxy
    Copyright (C) 2000 James Hollingshead

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

  cache.c -- Generic code for caching.

Overview:

   cache_init() redirects USER in ftp_cmds to come here. If user is
   anonymous we redirect all the other stuff we are going to need to
   parse or intercept for caching to come here too.

   We intercept RETR commands, and get all the info we need about the
   filename (size, last modification date, and uri).

   We call the specific cache code (in squidcache.c or localcache.c)
   with all this information, and it has the opportunity to return a
   file descriptor which will return the file.

   Specific cache code gets called with all incoming data in case it
   wishes to either alter it (ie. strip HTTP headers), or store it.

  Current Problems include: 
      o STAT does nothing useful.
      o If we start doing anything with USER or 220 cmds then we have
        the potential to interact with non-transparent proxying code.

  ***************************************/


#include <sys/signal.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "common.h"
#include "control.h"
#include "ftp-cmds.h"
#include "cache.h"

void cache_user(const char *cmd,const char *arg);
void cache_stat(const char *cmd,const char *arg);
void cache_abor(const char *cmd,const char *arg);
void cache_mode(const char *cmd,const char *arg);
void cache_stru(const char *cmd,const char *arg);
void cache_type(const char *cmd,const char *arg);
void cache_rest(const char *cmd,const char *arg);
void cache_retr(const char *cmd,const char *arg);

extern void abor_parse(const char *cmd, const char * arg);

static enum {
	NOTHING=FALSE,
	GOT_SIZE,
	GOT_MDTM,
	GOT_CWD,
	CLOSE
} expecting = NOTHING;

static int user_ind, mode_ind;
static int stru_ind, type_ind, rest_ind, retr_ind;

static int nooping=FALSE;
static int type=0; /*Ascii or Binary*/
static int noop=0; /*No of NOOPs sent*/
static time_t last_noop;
static int offset=0;

static struct {
	char host[MAX_LINE_LEN];
	char path[MAX_LINE_LEN];
	char filename[MAX_LINE_LEN];
	int size;
	char mdtm[MAX_LINE_LEN];
} fileinfo;

#define NOOP_INTERVAL 30 /*Seconds*/

static cmd_struct xfer_list[]  = {
	{ "ABOR", cache_abor },
	{ "STAT", cache_stat },
	{ "", NULL}
};
cmd_struct *main_cmdlist;

static int (*geninit)(char **cacheopts, char *chroot);
static int (*retr_start)(const char *host, const char *file, const char *mdtm,
               int size, int offset);
void (*inc_data)(char *inc, int *len);
static void (*retr_end)(void);
 

/* ------------------------------------------------------------- **
** Set up the function pointers to the cache module specified in
** cachemod. cacheopts contains options to be parsed by that module,
** and chroot is the directory we have been chrooted to --- the local
** cache module needs this to strip this part of the path out of its
** cachedir option. 
** ------------------------------------------------------------- */
int cache_geninit(char * cachemod, char **cacheopts, char *chroot)
{
	if(!strcasecmp(cachemod, "Local")){
		geninit=l_geninit;
		retr_start=l_retr_start;
		inc_data=l_inc_data;
		retr_end=l_retr_end;
	}
	else if(!strcasecmp(cachemod, "Squid")){
		geninit=s_geninit;
		retr_start=s_retr_start;
		inc_data=s_inc_data;
		retr_end=s_retr_end;
	}
	else{
		write_log(ERROR, "Unrecognised cache module %s", cachemod);
		return(-1);
	}

	if(geninit==NULL){
		write_log(ERROR, "Cache module %s is not compiled in", cachemod);
		return(-1);
	}

	return(geninit(cacheopts, chroot));
}

/* ------------------------------------------------------------- **
** Find and store the ftp_cmds that we may need to modify.
** Modify the ftp_cmds structure so that we get called on login.
** ------------------------------------------------------------- */
void cache_init(void)
{
	int i;

	for(i=0; *(ftp_cmds+i)->name!=0; i++){
		if(!strcasecmp((ftp_cmds+i)->name, "USER")) user_ind=i;
		else if(!strcasecmp((ftp_cmds+i)->name, "MODE")) mode_ind=i;
		else if(!strcasecmp((ftp_cmds+i)->name, "STRU")) stru_ind=i;
		else if(!strcasecmp((ftp_cmds+i)->name, "TYPE")) type_ind=i;
		else if(!strcasecmp((ftp_cmds+i)->name, "REST")) rest_ind=i;
		else if(!strcasecmp((ftp_cmds+i)->name, "RETR")) retr_ind=i;
	}

	(ftp_cmds+user_ind)->cmd=cache_user;
	main_cmdlist=ftp_cmds;
}

/* ------------------------------------------------------------- **
** If login is anonymous then redirect the cmds we need to intercept
** ------------------------------------------------------------- */
void cache_user(const char *cmd, const char *arg)
{
	send_command(cmd, arg);

	if(strncasecmp(arg,"ftp",3) && strncasecmp(arg,"anonymous",9))
		return;

	(ftp_cmds+mode_ind)->cmd=cache_mode;
	(ftp_cmds+stru_ind)->cmd=cache_stru;
	(ftp_cmds+type_ind)->cmd=cache_type;
	(ftp_cmds+rest_ind)->cmd=cache_rest;
	(ftp_cmds+retr_ind)->cmd=cache_retr;
}

/* ------------------------------------------------------------- **
** Commands to intercept at all times. FIXME Consider moving stru and
** mode functions to ftp-cmds.c They are pretty much obsolete, and
** probably not used by anyone any more. 
** ------------------------------------------------------------- */
void cache_mode(const char *cmd,const char *arg)
{
	if(!strncasecmp(arg, "S", 1)) send_message(200, "Command okay");
	else send_message(504,"Only stream mode implemented");
}

void cache_stru(const char *cmd,const char *arg)
{
	if(!strncasecmp(arg, "F", 1)) send_message(200, "Command okay");
	else send_message(504,"Only file structure implemented");
}

void cache_type(const char *cmd,const char *arg)
{
	if(!strncasecmp(arg, "I", 1)) {
		type=1;
		send_command(cmd,arg);
	}
	else if(!strncasecmp(arg, "AN", 1) || !strncasecmp(arg,"A",1)){
		type=0;
		send_command(cmd,arg);
	}
	else send_message(504,"Only types I and AN implemented");
}

void cache_rest(const char *cmd,const char *arg)
{

	debug("cache.c intercepted REST\n");
	offset=atoi(arg);

	send_command(cmd, arg);
	/*send_message(350,"Setting restart point. Send RETR to initiate transfer.");*/
}

/* ------------------------------------------------------------- **
** Commands to intercept during transfer
** ------------------------------------------------------------- */
void cache_stat(const char *cmd,const char *arg)
{
	send_message(213, "Retrieving file through cache");
}

void cache_abor(const char *cmd,const char *arg)
{
	debug("Closing both data connections\n");
	close(info->client_data.fd);
	close(info->server_data.fd);
	info->server_data.fd = info->client_data.fd = -1;
	info->state=NEITHER;

	/*FIXME call cache_end*/
	nooping=FALSE; /*FIXME clear up any remaining NOOPs*/

	debug("cache.c intercepted ABOR\n");
	send_message(426, "transfer aborted");
	send_message(226, "Closing data connection");
}

/* ------------------------------------------------------------- **
** The important bit. Intercept RETR commands, find out all we need to
** about the file, and then pass all the info to the caching function. 
** ------------------------------------------------------------- */
void cache_retr(const char *cmd,const char *arg)
{
	int i, code;
	char *tp, cwd[MAX_LINE_LEN-1], msg[MAX_LINE_LEN];

	if(type!=1) { 
		xfer_command(cmd, arg);		
		return; /*FIXME*/
	}

	strcpy(fileinfo.filename, arg);

	send_command("SIZE", arg);
	get_message(&code, msg);
	fileinfo.size=atoi(msg);
	debug2("Cache: Filesize is %d\n", fileinfo.size);

	send_command("MDTM",fileinfo.filename);
	get_message(&code, msg);
	if(code/100 != 2) *fileinfo.mdtm=0;
	else strcpy(fileinfo.mdtm, msg);
	debug2("Cache: MDTM is %s\n", fileinfo.mdtm);

	if(fileinfo.filename[0]!='/') {
		send_command("PWD", "");
		get_message(&code, msg);
		tp=strchr(msg,'"');
		if(tp==NULL) quit(SERVER_RUBBISH); /*FIXME a bit harsh*/
		strcpy(cwd,tp+1);
		tp=strchr(cwd,'"');
		if(tp==NULL) quit(SERVER_RUBBISH); /*FIXME a bit harsh*/
		if(*(tp-1)!='/') *(tp++)='/';
		*tp=0;

		strcpy(fileinfo.path, cwd);
		fileinfo.path[MAX_LINE_LEN-1]=0;
	} else {
		fileinfo.path[0]='\0';
	}

	strncpy(fileinfo.host, 
		inet_ntoa(info->server_control.address.sin_addr), MAX_LINE_LEN-1);
	fileinfo.host[MAX_LINE_LEN-1]=0;
	/*FIXME Append port if non default*/
	
	strncat(fileinfo.path, fileinfo.filename, MAX_LINE_LEN-strlen(fileinfo.path)-1);
	expecting=NOTHING;
	info->state=DOWNLOAD;

	i=retr_start(fileinfo.host, fileinfo.path, fileinfo.mdtm, fileinfo.size, offset);
	if(i==-1) {
		/*We have to do it ourselves*/
		nooping=FALSE;
		if(offset!=0) { /*Send another REST since we sent loads
				* of rubbish since the last one.*/
			sprintf(msg, "%d", offset);
			offset=0;
			send_command("REST", msg);
			get_message(&code, msg);
			if(code!=350) {
				send_message(503, "Can't do REST after all!");
				return;
			}
		}
		xfer_command("RETR", fileinfo.filename);
		return;
	} else if(offset!=0) { /*Lets reset the REST just in case*/
		send_command("REST", "0");
		get_message(&code, msg);
		offset=0;
	}

	/*Set up everything as if this were a normal connection*/
	if(info->server_data.fd != -1){
		close(info->server_data.fd);
		info->server_data.fd = -1;
	}
	info->server_data.fd=i;
	info->client_data.buflen=info->server_data.buflen=0;
	if( info->client_data.fd==-1 && info->mode != PASSIVE){
		if(config_transdata()){
			struct sockaddr_in tmp=info->server_control.address;
			tmp.sin_port=htons(20);
			info->client_data.fd = 
				transp_connect(info->client_data.address, 
					       tmp);
		}
		else info->client_data.fd =
			     connect_to_socket(info->client_data.address, 
					       ACTV);
	}
	
	fcntl(info->client_data.fd, F_SETFL, O_NONBLOCK);
	fcntl(info->server_data.fd, F_SETFL, O_NONBLOCK);	

	main_cmdlist=ftp_cmds;
	ftp_cmds=xfer_list;

	nooping=TRUE;
	time(&last_noop);
}

/* ------------------------------------------------------------- **
** Deal with server replies if we need to. 
** ------------------------------------------------------------- */
int cache_parsed_reply(int code, const char *msg)
{
	if(!config_usecache()) return(FALSE);
	if(noop>0){
		if(code>0) noop--;
		debug("Got NOOP reply\n");
		return(TRUE);
	}
	return(FALSE);
}		

void cache_inc_data(char *buf, int *len)
{
	if(!config_usecache()) return;

	if(nooping && time(NULL) - last_noop > NOOP_INTERVAL) {
		write(info->server_control.fd,"NOOP\r\n",6);
		noop++;
		time(&last_noop);
		debug("Sent NOOP\n");
	}
	inc_data(buf, len);
}

void cache_close_data(void)
{
	int code;
	char msg[MAX_LINE_LEN];

	if(!config_usecache()) return;

	while(noop>0) { 
		get_message(&code, msg);
		noop--;
	}

	nooping=FALSE;
        ftp_cmds=main_cmdlist;

	retr_end();
}
