/***************************************

    This is part of frox: A simple transparent FTP proxy
    Copyright (C) 2000 James Hollingshead

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

  control.c -- forward and parse the control stream.
  
  ***************************************/


#include <sys/ioctl.h>
#include <fcntl.h>
#include <errno.h>
#include <ctype.h>

#include "control.h"
#include "common.h"
#include "ftp-cmds.h"
#include "ftp-proxy.h"

int extract_clientcmd(char * buf, char * cmd, char * arg);
int extract_servercmd(char * buf, int * code, char * msg);
void do_linecontrol(char *buf, int reply);

/* ------------------------------------------------------------- **
** Forward buffered control stream data from client -> server.
** ------------------------------------------------------------- */
void client_control_forward()
{
        int i;
        char cmd[5], arg[MAX_LINE_LEN];

        while( (i=extract_clientcmd(info->client_control.buf, cmd, arg)) == 0 )
                parse_client_cmd(cmd, arg);

        if(i<0) {
                send_message(500,"You're sending me rubbish. Goodbye!");
                quit(CLIENT_RUBBISH);
        }
}

/* ------------------------------------------------------------- **
** Forward buffered control stream data from server -> client.
** ------------------------------------------------------------- */
void server_control_forward()
{
        int i, code;
        char arg[MAX_LINE_LEN];

        while( (i=extract_servercmd(info->server_control.buf, &code, arg)) == 0 )
                if(!cache_parsed_reply(code, arg)) send_message(code,arg);

        if(i<0) {
                send_message(500,"FTP server is sending you rubbish! Closing connection.");
                quit(SERVER_RUBBISH);
        }
}

void parse_client_cmd(char *cmd, char *arg)
{
	cmd_struct *p;

	for(p=ftp_cmds; *p->name!=0; p++){
		if(!strcasecmp(p->name, cmd)){
			p->cmd(cmd, arg);
			return;
		}
	}
	debug2("Command %s not implemented\n", cmd);
	send_message(502, "Command not implemented.");
}

/* ------------------------------------------------------------- **
** Get one command from client.
** ------------------------------------------------------------- */
void get_command(char *cmd, char *arg )
{
	int i;

	while( (i=extract_clientcmd(info->client_control.buf, cmd, arg)) ==1)
		if(get_control_line(GET_CLNT)<=0) quit(SOCKET);
}

/* ------------------------------------------------------------- **
** Get one reply from server.
** ------------------------------------------------------------- */
void get_message(int *code, char *msg )
{
	int i;

	while( (i=extract_servercmd(info->server_control.buf, code, msg)) == 1)
		if(get_control_line(GET_SRVR)<=0) quit(SOCKET);
}


/* ------------------------------------------------------------- **
** Send command to server 
** ------------------------------------------------------------- */
void send_command(const char * cmd, const char *arg )
{
	char buf[BUF_LEN];

	if(strlen(arg)==0)
		snprintf(buf,BUF_LEN-1, "%s\r\n",cmd);
	else    snprintf(buf,BUF_LEN-1,"%s %s\r\n",cmd, arg);

	write_all(info->server_control.fd, buf);

	debug2("  C: \033[31m%s\033[37m",buf); /*Ooooh Colour*/
}
		
/* ------------------------------------------------------------- **
** Send message to client
** ------------------------------------------------------------- */
void send_message(int code, const char *msg)
{
	char buf[BUF_LEN];

	if(code>0) snprintf(buf,BUF_LEN-1,"%d %s\r\n",code, msg);
	else if(code<0) snprintf(buf,BUF_LEN-1,"%d-%s\r\n",-code, msg); 
	else snprintf(buf, BUF_LEN-1, "%s\r\n", msg);

	write_all(info->client_control.fd, buf);

	debug2("  S: \033[34m%s\033[37m",buf);	
}

/* ------------------------------------------------------------- **
** Central select bit. Deals with data connection
** forwarding/listening, and quits on ctrl connection close. Once
** there is a complete line read from one of the control connctions
** specified in "which" (GET_SRVR, GET_CTRL, or GET_SRVR|GET_CTRL) we
** return.
**
** The line read into the control connection buffer on function return
** contains a newline (\n), and is NULL terminated at some point
** beyond that. No other checking has been done on it.
**
** Return value is one of GET_SRVR, GET_CTRL or GET_SRVR|GET_CTRL.
** ------------------------------------------------------------- */
int get_control_line(int which)
{
	int ret=0, i;
	fd_set reads, writes;
	struct timeval timeout;

	do{
		setup_fds(&reads, &writes, &timeout);
		
		if(iselect(FD_SETSIZE, &reads, &writes, NULL, &timeout)==0)
			quit(TIMEOUT);

		do_dataforward(&reads,&writes);		
		
		if(FD_ISSET(info->client_control.fd,&reads)) {
			i=append_read(info->client_control.fd, info->client_control.buf);
			if(i==0) quit(CLIENT_CLOSE);
			if(i<0) quit(CLIENT_RUBBISH);/*FIXME just stop selecting*/
			if(strchr(info->client_control.buf, '\n')) ret |= GET_CLNT;
		}
		
		if(info->server_control.fd!=-1 &&
		   FD_ISSET(info->server_control.fd,&reads)){
			i=append_read(info->server_control.fd, info->server_control.buf);
			if(i==0) quit(SERVER_CLOSE);
			if(i<0) quit(SERVER_RUBBISH);/*FIXME just stop selecting*/
			if(strchr(info->server_control.buf, '\n')) ret |= GET_SRVR;
		}

	}while(! (ret & which));
	return(ret);
}

/* ------------------------------------------------------------- **
** Process buf to remove telnet line control stuff.
** ------------------------------------------------------------- */
void do_linecontrol(char *buf, int reply)
{
	char *p;
	int c, len;

	while( (p=strchr(buf, IAC)) != NULL){
		c=p[1] & 255;
		len=strlen(p);
		switch(c) {
		case WILL:
			if(len<3) quit(CLIENT_RUBBISH);
			p[1]=DONT;
			send(reply, p, 3, MSG_OOB);
			memmove(p, p+3, len-2);
			break;
		case DO:
			if(len<3) quit(CLIENT_RUBBISH);
			p[1]=WONT;
			send(reply, p, 3, MSG_OOB);
			memmove(p, p+3, len-2);
			break;
		case WONT:
		case DONT:
			if(len<3) quit(CLIENT_RUBBISH);
			memmove(p, p+3, len-2);
			break;
		case IAC:
			if(len<1) quit(CLIENT_RUBBISH);
			memmove(p, p+1, len);
			break;
		case IP:
		case DM:
			if(len<2) quit(CLIENT_RUBBISH);
			memmove(buf, p+2, len-1);
			break;
		default:
			if(len<2) quit(CLIENT_RUBBISH);
			memmove(p, p+2, len-1);
			break;
		}
	}
}

/***************************************************************************
 *
 * Functions which read the raw control stream --- be careful
 *
 **************************************************************************/

/* ------------------------------------------------------------- **
** Removes one line of control stream from buf (up to '\r\n'). <= 5
** chars are returned in cmd (up to 4 + \0), and <= MAX_LINE_LEN in
** arg. Must accept any NULL terminated buf and give sane output.
**  
** returns 0 on success, 1 if there is not a complete line in
** buf, and -X on non-sane buf. contents of buf unchanged on
** return(1), undefined on return(-X)
** ------------------------------------------------------------- */
int extract_clientcmd(char *buf, char *cmd, char *arg)
{
	int i;
	char *p,*cmdend,*argstart,*end;

	do_linecontrol(buf, info->client_control.fd);

	cmdend=argstart=end=p=strpbrk(buf,"\r\n \t");
	if(p==NULL) return(1);

	if(*p==' ' || *p=='\t'){
		for(argstart=p;*argstart==' ' || *argstart=='\t';argstart++);
		end=p=strpbrk(p, "\r\n");
	}
	if(p==NULL) return(1);
	if(*p=='\r')
		if( (end=strpbrk(p,"\n"))==NULL ) return(1);
	
	/* We have a complete command. Start copying it over.
	   eg. "USER   ftp\r\r\r\n"  (Microsoft servers don't do \r\n !)
	   buf__|   |  |  |     |__end
        cmdend______|  |  |______p
	argstart_______|               */

	*cmdend=0;
	*p=0;

	if(strlen(buf)>4 || strlen(argstart) > MAX_LINE_LEN-1 ) {
		debug2("Client sent us \"%s\", ",buf);
		debug2("\"%s\"\n",argstart);
		return(-1);
	}
	strcpy(cmd, buf);
	strcpy(arg,argstart);
       	memmove(buf, end+1, strlen(end+1)+1);

	for(i=0;i<strlen(arg);i++)
		if(arg[i]>126 || arg[i]<32) arg[i]='?';

	return(0);
}

/* ------------------------------------------------------------- **
** As extract_clientcmd, but returns the code as an int.
** If code==0 then we are in a multiline. code<0 means we are
** starting a multiline. 
** ------------------------------------------------------------- */
int extract_servercmd(char *buf, int * code, char *msg)
{
	static int multiline=0;
	int i;
	char *p,*end;

	do_linecontrol(buf, info->client_control.fd);

	end=p=strpbrk(buf,"\r\n");

	if(p==NULL) return(1);
	if(*p=='\r')
		if( (end=strpbrk(p,"\n"))==NULL ) return(1);
	
	/* We have a complete command. Start copying it over.
	   eg. "501  go away now\r\r\n"  (Microsoft servers don't do \r\n !)
	   buf__|               |   |__end
	                        |______p       */

	*p=0;

	if(multiline == 0){
		for(i=0;i<3;i++) if( buf[i] < '0' || buf[i] > '9') return(-1);
		if(buf[3]!=' ' && buf[3]!='-') return(-1);
		*code=atoi(buf);
		if(buf[3] == '-') {  /* Start of a multiline */

			multiline = *code;
			*code=-*code;
		}
		if(strlen(buf+4) > MAX_LINE_LEN-1 ) return(-1);
		strcpy(msg,buf+4);
	} else {
		if(isdigit(buf[0]) && isdigit(buf[1]) && isdigit(buf[2])
		   && buf[3]==' ' && (*code=atoi(buf)) == multiline ) { /* End multiline */
			multiline=0;
			if(strlen(buf+4) > MAX_LINE_LEN-1) return(-1);
			strcpy(msg,buf+4);
		} else {             /* Continue Multiline */
			*code=0;
			if(strlen(buf) > MAX_LINE_LEN-1) return(-1);
			strcpy(msg,buf);
		}
	}
	memmove(buf, end+1, strlen(end+1)+1);

	for(i=0;i<strlen(msg);i++)
		if(msg[i]>126 || msg[i]<32) msg[i]='?';

	return(0);
}

/* ------------------------------------------------------------- **
** Write buf->fd. If fd has O_NONBLOCK set we will keep trying until
** TIMEOUT. ?FIXME do away with this and just set alarm() and die on
** SIGALRM
** ------------------------------------------------------------- */
void write_all(int fd, char *buf)
{
        fd_set writes;
        struct timeval timeout;
        char *p=buf;
        int i=0, tmp;
        time_t end;

        end=time(NULL)+config_timeout().tv_sec; /*Should we use another value?*/

        do{
                FD_ZERO(&writes);
                FD_SET(fd, &writes);
                if( (tmp=end-time(NULL))<0) break;
                timeout.tv_sec=tmp;
                timeout.tv_usec=0;

                i=iselect(FD_SETSIZE, NULL, &writes, NULL, &timeout);
                if(i<1) break;
                i=iwrite(fd, p,strlen(p));
                if(i<1) break;
                p+=i;
        } while(i<strlen(p));

        if(i<1){
                perror("writeall");
                write_log(ERROR, "Error writing to socket. Exiting");
                closeandexit(-1);
        }

}
